/**
 * Set of logical utilities to deal with page global behaviour
 * @param {object} wizards collection of selector / callback on which
 * the functions called will be performed (promises are supported)
 * 
 * @Feature (name=magic)
 */
function MagicTrick(...wizards) {

    const that = this;

    // Initializing DOM access and removing unusable data
    for (const wizard of wizards) {
        let processable = (wizard.selector && wizard.callback);
        if (processable) {
            wizard.entity = $(wizard.selector);
            if (!wizard.entity.length) {
                processable = false;
            }
        }
        if (!processable) {
            wizards = wizards.filter(data => data !== wizard);
            console.error(`MagicTrick: ${wizard} is unreachable or does not contain the properties { selector, callback }`);
        }
    }

    /**
     * Performs callbacks when revealed by scrolling
     */
    const reveal = () => {

        const sleepingWizards = [];
        $(window).scroll(function() {

            if (wizards.length !== sleepingWizards.length) {
                const pos = $(window).scrollTop() + $(window).height();
                for (const wizard of wizards) {
                    if (pos >= wizard.entity.scrollTop() && !sleepingWizards.includes(wizard)) {
                        wizard.callback(wizard.entity);
                        sleepingWizards.push(wizard);
                    }
                }   
            }
        });
        return that;
    }

    /**
     * Builds wrapper object to execute all callbacks 
     * @returns new promise
     */
    const factorize = () => {
        
        let asyncCount = 1;
        const asyncChilds = [];
        return new Promise((resolve, reject) => {
            for (const wizard of wizards) {
                if (wizard.callback instanceof Promise) {
                    asyncChilds.push(wizard.callback);
                } else {
                    wizard.callback(wizard.entity);
                }
            }
            asyncChilds.forEach(callback => 
                callback(wizard.entity).then(() => asyncCount++).catch(reject));
            while (asyncCount !== asyncChilds.length) {
                // Threads synchronization
            }
            resolve(wizards);
        });
    }

    /**
     * Performs callbacks one by one with sleep time
     * @param {number} delay = sleep time ms value (default: 200)
     */
    const defeer = (delay) => {

        let count = 0;
        if (!delay) delay = 200;
        delay = parseInt(delay);
        for (const wizard of wizards) {
            count++;
            setTimeout(() => wizard.callback(wizard.entity), delay * count);
        }
        return that;
    }

    return {
        defeer: defeer,
        factorize: factorize,
        reveal: reveal
    }
}